Make idev, hlink and file_struct + strings use allocation
[rsync/rsync.git] / hlink.c
diff --git a/hlink.c b/hlink.c
index 40a7aa9..f42df53 100644 (file)
--- a/hlink.c
+++ b/hlink.c
@@ -38,34 +38,49 @@ static int hlink_compare(struct file_struct **file1, struct file_struct **file2)
        return file_compare(file1, file2);
 }
 
-static struct file_struct **hlink_list;
-static int hlink_count;
+struct file_struct **hlink_list;
+int hlink_count;
 
 #define LINKED(p1,p2) ((p1)->F_DEV == (p2)->F_DEV \
                    && (p1)->F_INODE == (p2)->F_INODE)
 
 /* Analyze the data in the hlink_list[], remove items that aren't multiply
- * linked, and replace the dev+inode data with the head+next linked list. */
-static void link_idev_data(void)
+ * linked, and replace the dev+inode data with the hlindex+next linked list. */
+static void link_idev_data(struct file_list *flist)
 {
        struct file_struct *head;
        int from, to, start;
 
+       alloc_pool_t hlink_pool;
+       alloc_pool_t idev_pool = flist->hlink_pool;
+
+       hlink_pool = pool_create(128 * 1024, sizeof (struct hlink),
+           out_of_memory, POOL_INTERN);
+
        for (from = to = 0; from < hlink_count; from++) {
                start = from;
                head = hlink_list[start];
                while (from < hlink_count-1
                    && LINKED(hlink_list[from], hlink_list[from+1])) {
-                       hlink_list[from]->F_HEAD = head;
+                       pool_free(idev_pool, 0, hlink_list[from]->link_u.idev);
+                       hlink_list[from]->link_u.links = pool_talloc(hlink_pool,
+                           struct hlink, 1, "hlink_list");
+
+                       hlink_list[from]->F_HLINDEX = to;
                        hlink_list[from]->F_NEXT = hlink_list[from+1];
                        from++;
                }
                if (from > start) {
-                       hlink_list[from]->F_HEAD = head;
-                       hlink_list[from]->F_NEXT = NULL;
+                       pool_free(idev_pool, 0, hlink_list[from]->link_u.idev);
+                       hlink_list[from]->link_u.links = pool_talloc(hlink_pool,
+                           struct hlink, 1, "hlink_list");
+
+                       hlink_list[from]->F_HLINDEX = to;
+                       hlink_list[from]->F_NEXT = head;
+                       hlink_list[from]->flags |= FLAG_HLINK_EOL;
                        hlink_list[to++] = head;
                } else {
-                       free((char*)head->link_u.idev);
+                       pool_free(idev_pool, 0, head->link_u.idev);
                        head->link_u.idev = NULL;
                }
        }
@@ -73,12 +88,16 @@ static void link_idev_data(void)
        if (!to) {
                free(hlink_list);
                hlink_list = NULL;
+               pool_destroy(hlink_pool);
+               hlink_pool = NULL;
        } else {
                hlink_count = to;
                if (!(hlink_list = realloc_array(hlink_list,
-                                       struct file_struct *, hlink_count)))
+                   struct file_struct *, hlink_count)))
                        out_of_memory("init_hard_links");
        }
+       flist->hlink_pool = hlink_pool;
+       pool_destroy(idev_pool);
 }
 #endif
 
@@ -103,26 +122,42 @@ void init_hard_links(struct file_list *flist)
        }
 
        qsort(hlink_list, hlink_count,
-             sizeof(hlink_list[0]), (int (*)()) hlink_compare);
+           sizeof hlink_list[0], (int (*)()) hlink_compare);
 
        if (!hlink_count) {
                free(hlink_list);
                hlink_list = NULL;
        } else
-               link_idev_data();
+               link_idev_data(flist);
 #endif
 }
 
+int hard_link_check(struct file_struct *file, int skip)
+{
+       if (!file->link_u.links)
+               return 0;
+       if (skip && !(file->flags & FLAG_HLINK_EOL))
+               hlink_list[file->F_HLINDEX] = file->F_NEXT;
+       if (hlink_list[file->F_HLINDEX] != file) {
+               if (verbose > 1) {
+                       rprintf(FINFO, "\"%s\" is a hard link\n",
+                           f_name(file));
+               }
+               return 1;
+       }
+       return 0;
+}
+
 #if SUPPORT_HARD_LINKS
 static void hard_link_one(char *hlink1, char *hlink2)
 {
        if (do_link(hlink1, hlink2)) {
-               if (verbose > 0) {
+               if (verbose) {
                        rprintf(FINFO, "link %s => %s failed: %s\n",
-                               hlink2, hlink1, strerror(errno));
+                           hlink2, hlink1, strerror(errno));
                }
        }
-       else if (verbose > 0)
+       else if (verbose)
                rprintf(FINFO, "%s => %s\n", hlink2, hlink1);
 }
 #endif
@@ -136,9 +171,9 @@ static void hard_link_one(char *hlink1, char *hlink2)
 void do_hard_links(void)
 {
 #if SUPPORT_HARD_LINKS
-       struct file_struct *file;
-       char fbuf[MAXPATHLEN];
-       char *hlink1, *hlink2;
+       struct file_struct *file, *first;
+       char hlink1[MAXPATHLEN];
+       char *hlink2;
        STRUCT_STAT st1, st2;
        int i;
 
@@ -146,11 +181,10 @@ void do_hard_links(void)
                return;
 
        for (i = 0; i < hlink_count; i++) {
-               file = hlink_list[i];
-               hlink1 = f_name_to(file, fbuf, sizeof fbuf);
-               if (link_stat(hlink1, &st1) != 0)
+               first = file = hlink_list[i];
+               if (link_stat(f_name_to(first, hlink1), &st1) != 0)
                        continue;
-               while ((file = file->F_NEXT) != NULL) {
+               while ((file = file->F_NEXT) != first) {
                        hlink2 = f_name(file);
                        if (link_stat(hlink2, &st2) == 0) {
                                if (st2.st_dev == st1.st_dev
@@ -159,9 +193,9 @@ void do_hard_links(void)
                                if (robust_unlink(hlink2)) {
                                        if (verbose > 0) {
                                                rprintf(FINFO,
-                                                       "unlink %s failed: %s\n",
-                                                       full_fname(hlink2), 
-                                                       strerror(errno));
+                                                   "unlink %s failed: %s\n",
+                                                   full_fname(hlink2), 
+                                                   strerror(errno));
                                        }
                                        continue;
                                }